@@ -17,35 +17,35 @@ class AbstractItemEncoder(object):
     supportIndefLenMode = 1
 
     # noinspection PyMethodMayBeStatic
-    def encodeTag(self, t, isConstructed):
-        tagClass, tagFormat, tagId = t.asTuple()  # this is a hotspot
-        v = tagClass | tagFormat
+    def encodeTag(self, singleTag, isConstructed):
+        tagClass, tagFormat, tagId = singleTag
+        encodedTag = tagClass | tagFormat
         if isConstructed:
-            v |= tag.tagFormatConstructed
+            encodedTag |= tag.tagFormatConstructed
         if tagId < 31:
-            return int2oct(v | tagId)
+            return (encodedTag | tagId,)
         else:
-            s = int2oct(tagId & 0x7f)
+            substrate = (tagId & 0x7f,)
             tagId >>= 7
             while tagId:
-                s = int2oct(0x80 | (tagId & 0x7f)) + s
+                substrate = (0x80 | (tagId & 0x7f),) + substrate
                 tagId >>= 7
-            return int2oct(v | 0x1F) + s
+            return (encodedTag | 0x1F,) + substrate
 
     def encodeLength(self, length, defMode):
         if not defMode and self.supportIndefLenMode:
-            return int2oct(0x80)
+            return (0x80,)
         if length < 0x80:
-            return int2oct(length)
+            return (length,)
         else:
-            substrate = null
+            substrate = ()
             while length:
-                substrate = int2oct(length & 0xff) + substrate
+                substrate = (length & 0xff,) + substrate
                 length >>= 8
             substrateLen = len(substrate)
             if substrateLen > 126:
                 raise error.PyAsn1Error('Length octets overflow (%d)' % substrateLen)
-            return int2oct(0x80 | substrateLen) + substrate
+            return (0x80 | substrateLen,) + substrate
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
         raise error.PyAsn1Error('Not implemented')
@@ -57,89 +57,94 @@ class AbstractItemEncoder(object):
             return encodeFun(eoo.endOfOctets, defMode)
 
     def encode(self, encodeFun, value, defMode, maxChunkSize):
-        substrate, isConstructed = self.encodeValue(
+        substrate, isConstructed, isOctets = self.encodeValue(
             encodeFun, value, defMode, maxChunkSize
         )
-        tagSet = value.getTagSet()
+        tagSet = value.tagSet
+        # tagged value?
         if tagSet:
             if not isConstructed:  # primitive form implies definite mode
                 defMode = True
-            return self.encodeTag(
-                tagSet[-1], isConstructed
-            ) + self.encodeLength(
-                len(substrate), defMode
-            ) + substrate + self._encodeEndOfOctets(encodeFun, defMode)
-        else:
-            return substrate  # untagged value
+            header = self.encodeTag(tagSet[-1], isConstructed)
+            header += self.encodeLength(len(substrate), defMode)
+
+            if isOctets:
+                substrate = ints2octs(header) + substrate
+            else:
+                substrate = ints2octs(header + substrate)
+
+            eoo =  self._encodeEndOfOctets(encodeFun, defMode)
+            if eoo:
+                substrate += eoo
+
+        return substrate
 
 
 class EndOfOctetsEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        return null, 0
+        return null, False, True
 
 
 class ExplicitlyTaggedItemEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
         if isinstance(value, base.AbstractConstructedAsn1Item):
-            value = value.clone(tagSet=value.getTagSet()[:-1],
-                                cloneValueFlag=1)
+            value = value.clone(tagSet=value.tagSet[:-1], cloneValueFlag=True)
         else:
-            value = value.clone(tagSet=value.getTagSet()[:-1])
-        return encodeFun(value, defMode, maxChunkSize), 1
+            value = value.clone(tagSet=value.tagSet[:-1])
+        return encodeFun(value, defMode, maxChunkSize), True, True
 
 
 explicitlyTaggedItemEncoder = ExplicitlyTaggedItemEncoder()
 
 
 class BooleanEncoder(AbstractItemEncoder):
-    supportIndefLenMode = 0
-    _true = ints2octs((1,))
-    _false = ints2octs((0,))
+    supportIndefLenMode = False
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        return value and self._true or self._false, 0
+        return value and (1,) or (0,), False, False
 
 
 class IntegerEncoder(AbstractItemEncoder):
-    supportIndefLenMode = 0
+    supportIndefLenMode = False
     supportCompactZero = False
-    encodedZero = ints2octs((0,))
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
         if value == 0:
             # de-facto way to encode zero
             if self.supportCompactZero:
-                return null, 0
+                return (), False, False
             else:
-                return self.encodedZero, 0
+                return (0,), False, False
 
-        return to_bytes(int(value), signed=True), 0
+        return to_bytes(int(value), signed=True), False, True
 
 
 class BitStringEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        if len(value) % 8:
-            alignedValue = value << (8 - len(value) % 8)
+        valueLength = len(value)
+        if valueLength % 8:
+            alignedValue = value << (8 - valueLength % 8)
         else:
             alignedValue = value
 
         if not maxChunkSize or len(alignedValue) <= maxChunkSize * 8:
             substrate = alignedValue.asOctets()
-            return int2oct(len(substrate) * 8 - len(value)) + substrate, 0
+            return int2oct(len(substrate) * 8 - valueLength) + substrate, False, True
 
         stop = 0
         substrate = null
-        while stop < len(value):
+        while stop < valueLength:
             start = stop
-            stop = min(start + maxChunkSize * 8, len(value))
+            stop = min(start + maxChunkSize * 8, valueLength)
             substrate += encodeFun(alignedValue[start:stop], defMode, maxChunkSize)
-        return substrate, 1
+
+        return substrate, True, True
 
 
 class OctetStringEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
         if not maxChunkSize or len(value) <= maxChunkSize:
-            return value.asOctets(), 0
+            return value.asOctets(), False, True
         else:
             pos = 0
             substrate = null
@@ -147,31 +152,33 @@ class OctetStringEncoder(AbstractItemEncoder):
                 v = value.clone(value[pos:pos + maxChunkSize])
                 if not v:
                     break
-                substrate = substrate + encodeFun(v, defMode, maxChunkSize)
+                substrate += encodeFun(v, defMode, maxChunkSize)
                 pos += maxChunkSize
-            return substrate, 1
+
+            return substrate, True, True
 
 
 class NullEncoder(AbstractItemEncoder):
-    supportIndefLenMode = 0
+    supportIndefLenMode = False
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        return null, 0
+        return null, False, True
 
 
 class ObjectIdentifierEncoder(AbstractItemEncoder):
-    supportIndefLenMode = 0
+    supportIndefLenMode = False
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
         oid = value.asTuple()
-        if len(oid) < 2:
-            raise error.PyAsn1Error('Short OID %s' % (value,))
-
-        octets = ()
 
         # Build the first pair
-        first = oid[0]
-        second = oid[1]
+        try:
+            first = oid[0]
+            second = oid[1]
+
+        except IndexError:
+            raise error.PyAsn1Error('Short OID %s' % (value,))
+
         if 0 <= second <= 39:
             if first == 1:
                 oid = (second + 40,) + oid[2:]
@@ -186,6 +193,8 @@ class ObjectIdentifierEncoder(AbstractItemEncoder):
         else:
             raise error.PyAsn1Error('Impossible first/second arcs at %s' % (value,))
 
+        octets = ()
+
         # Cycle through subIds
         for subOid in oid:
             if 0 <= subOid <= 127:
@@ -203,7 +212,7 @@ class ObjectIdentifierEncoder(AbstractItemEncoder):
             else:
                 raise error.PyAsn1Error('Negative OID arc %s at %s' % (subOid, value))
 
-        return ints2octs(octets), 0
+        return octets, False, False
 
 
 class RealEncoder(AbstractItemEncoder):
@@ -247,25 +256,26 @@ class RealEncoder(AbstractItemEncoder):
         encbase = 2
         e = float('inf')
         for i in range(3):
-            sign, mantissa[i], encBase[i], exponenta[i] = \
-                self._dropFloatingPoint(mantissa[i], encBase[i], exponenta[i])
-            if abs(exponenta[i]) < abs(e) or \
-                    (abs(exponenta[i]) == abs(e) and mantissa[i] < m):
+            (sign,
+             mantissa[i],
+             encBase[i],
+             exponenta[i]) = self._dropFloatingPoint(mantissa[i], encBase[i], exponenta[i])
+            if abs(exponenta[i]) < abs(e) or (abs(exponenta[i]) == abs(e) and mantissa[i] < m):
                 e = exponenta[i]
                 m = int(mantissa[i])
                 encbase = encBase[i]
         return sign, m, encbase, e
 
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        if value.isPlusInfinity():
-            return int2oct(0x40), 0
-        if value.isMinusInfinity():
-            return int2oct(0x41), 0
+        if value.isPlusInf:
+            return (0x40,), False, False
+        if value.isMinusInf:
+            return (0x41,), False, False
         m, b, e = value
         if not m:
-            return null, 0
+            return null, False, True
         if b == 10:
-            return str2octs('\x03%dE%s%d' % (m, e == 0 and '+' or '', e)), 0
+            return str2octs('\x03%dE%s%d' % (m, e == 0 and '+' or '', e)), False, True
         elif b == 2:
             fo = 0x80  # binary encoding
             ms, m, encbase, e = self._chooseEncBase(value)
@@ -321,28 +331,26 @@ class RealEncoder(AbstractItemEncoder):
                 po = int2oct(m & 0xff) + po
                 m >>= 8
             substrate = int2oct(fo) + eo + po
-            return substrate, 0
+            return substrate, False, True
         else:
             raise error.PyAsn1Error('Prohibited Real base %s' % b)
 
 
 class SequenceEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        value.setDefaultComponents()
         value.verifySizeSpec()
+        namedTypes = value.componentType
         substrate = null
         idx = len(value)
         while idx > 0:
             idx -= 1
-            if value[idx] is None:  # Optional component
-                continue
-            component = value.getDefaultComponentByPosition(idx)
-            if component is not None and component == value[idx]:
-                continue
-            substrate = encodeFun(
-                value[idx], defMode, maxChunkSize
-            ) + substrate
-        return substrate, 1
+            if namedTypes:
+                if namedTypes[idx].isOptional and not value[idx].isValue:
+                    continue
+                if namedTypes[idx].isDefaulted and value[idx] == namedTypes[idx].asn1Object:
+                    continue
+            substrate = encodeFun(value[idx], defMode, maxChunkSize) + substrate
+        return substrate, True, True
 
 
 class SequenceOfEncoder(AbstractItemEncoder):
@@ -352,20 +360,18 @@ class SequenceOfEncoder(AbstractItemEncoder):
         idx = len(value)
         while idx > 0:
             idx -= 1
-            substrate = encodeFun(
-                value[idx], defMode, maxChunkSize
-            ) + substrate
-        return substrate, 1
+            substrate = encodeFun(value[idx], defMode, maxChunkSize) + substrate
+        return substrate, True, True
 
 
 class ChoiceEncoder(AbstractItemEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        return encodeFun(value.getComponent(), defMode, maxChunkSize), 1
+        return encodeFun(value.getComponent(), defMode, maxChunkSize), True, True
 
 
 class AnyEncoder(OctetStringEncoder):
     def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
-        return value.asOctets(), defMode == 0
+        return value.asOctets(), defMode == False, True
 
 
 tagMap = {
@@ -400,14 +406,39 @@ tagMap = {
     useful.UTCTime.tagSet: OctetStringEncoder()
 }
 
-# Type-to-codec map for ambiguous ASN.1 types
+# Put in ambiguous & non-ambiguous types for faster codec lookup
 typeMap = {
+    univ.Boolean.typeId: BooleanEncoder(),
+    univ.Integer.typeId: IntegerEncoder(),
+    univ.BitString.typeId: BitStringEncoder(),
+    univ.OctetString.typeId: OctetStringEncoder(),
+    univ.Null.typeId: NullEncoder(),
+    univ.ObjectIdentifier.typeId: ObjectIdentifierEncoder(),
+    univ.Enumerated.typeId: IntegerEncoder(),
+    univ.Real.typeId: RealEncoder(),
+    # Sequence & Set have same tags as SequenceOf & SetOf
     univ.Set.typeId: SequenceEncoder(),
     univ.SetOf.typeId: SequenceOfEncoder(),
     univ.Sequence.typeId: SequenceEncoder(),
     univ.SequenceOf.typeId: SequenceOfEncoder(),
     univ.Choice.typeId: ChoiceEncoder(),
-    univ.Any.typeId: AnyEncoder()
+    univ.Any.typeId: AnyEncoder(),
+    # character string types
+    char.UTF8String.typeId: OctetStringEncoder(),
+    char.NumericString.typeId: OctetStringEncoder(),
+    char.PrintableString.typeId: OctetStringEncoder(),
+    char.TeletexString.typeId: OctetStringEncoder(),
+    char.VideotexString.typeId: OctetStringEncoder(),
+    char.IA5String.typeId: OctetStringEncoder(),
+    char.GraphicString.typeId: OctetStringEncoder(),
+    char.VisibleString.typeId: OctetStringEncoder(),
+    char.GeneralString.typeId: OctetStringEncoder(),
+    char.UniversalString.typeId: OctetStringEncoder(),
+    char.BMPString.typeId: OctetStringEncoder(),
+    # useful types
+    useful.ObjectDescriptor.typeId: OctetStringEncoder(),
+    useful.GeneralizedTime.typeId: OctetStringEncoder(),
+    useful.UTCTime.typeId: OctetStringEncoder()
 }
 
 
@@ -420,34 +451,34 @@ class Encoder(object):
         self.__typeMap = typeMap
 
     def __call__(self, value, defMode=True, maxChunkSize=0):
-        if not isinstance(value, base.Asn1Item):
-            raise error.PyAsn1Error('value is not valid (should be an instance of an ASN.1 Item)')
         if not defMode and not self.supportIndefLength:
             raise error.PyAsn1Error('Indefinite length encoding not supported by this codec')
-        debug.logger & debug.flagEncoder and debug.logger(
-            'encoder called in %sdef mode, chunk size %s for type %s, value:\n%s' % (
-                not defMode and 'in' or '', maxChunkSize, value.prettyPrintType(), value.prettyPrint()))
-        tagSet = value.getTagSet()
+        if debug.logger & debug.flagEncoder:
+            logger = debug.logger
+        else:
+            logger = None
+        if logger:
+            logger('encoder called in %sdef mode, chunk size %s for type %s, value:\n%s' % (not defMode and 'in' or '', maxChunkSize, value.prettyPrintType(), value.prettyPrint()))
+        tagSet = value.tagSet
         if len(tagSet) > 1:
             concreteEncoder = explicitlyTaggedItemEncoder
         else:
-            if value.typeId is not None and value.typeId in self.__typeMap:
+            try:
                 concreteEncoder = self.__typeMap[value.typeId]
-            elif tagSet in self.__tagMap:
-                concreteEncoder = self.__tagMap[tagSet]
-            else:
-                tagSet = value.baseTagSet
-                if tagSet in self.__tagMap:
-                    concreteEncoder = self.__tagMap[tagSet]
-                else:
+            except KeyError:
+                # use base type for codec lookup to recover untagged types
+                baseTagSet = tag.TagSet(value.tagSet.baseTag, value.tagSet.baseTag)
+                try:
+                    concreteEncoder = self.__tagMap[baseTagSet]
+                except KeyError:
                     raise error.PyAsn1Error('No encoder for %s' % (value,))
-        debug.logger & debug.flagEncoder and debug.logger(
-            'using value codec %s chosen by %s' % (concreteEncoder.__class__.__name__, tagSet))
+            if logger:
+                logger('using value codec %s chosen by %s' % (concreteEncoder.__class__.__name__, tagSet))
         substrate = concreteEncoder.encode(
             self, value, defMode, maxChunkSize
         )
-        debug.logger & debug.flagEncoder and debug.logger(
-            'built %s octets of substrate: %s\nencoder completed' % (len(substrate), debug.hexdump(substrate)))
+        if logger:
+            logger('built %s octets of substrate: %s\nencoder completed' % (len(substrate), debug.hexdump(substrate)))
         return substrate
 
 #: Turns ASN.1 object into BER octet stream.
